# Wishrc startup for ToolScript (irsim)
#
# For installation:  Put this file and also tclirsim.so into
# directory ${CAD_HOME}/lib/irsim/tcl/, and set the "load" line below
# to point to the location of tclirsim.so.  Also see comments
# in shell script "irsim.sh".
#

package require BLT

load TCL_DIR/tclirsim.so

# Load the "random" library, if it exists

catch {load TCL_DIR/random.so}

# It is important to make sure no irsim commands overlap with Tcl built-in
# commands, because otherwise the namespace import will fail.

proc pushnamespace { name } {

   set y [namespace eval ${name} info commands ::${name}::*]
   set z [info commands]

   foreach v $y {
      regsub -all {\*} $v {\\*} i
      set x [namespace tail $i]
      if {[lsearch $z $x] < 0} {
         namespace import $i
      } else {
            puts "Warning: ${name} command '$x' use fully-qualified name '$v'"
      }
   }
}

proc popnamespace { name } {
   set z [info commands]
   set l [expr [string length ${name}] + 5]

   while {[set v [lsearch $z ${name}_tcl_*]] >= 0} {
      set y [lindex $z $v]
      set w [string range $y $l end]
      interp alias {} ::$w {}
      rename ::$y ::$w
      puts "Info: replacing ::$w with ::$y"
   }
   namespace forget ::${name}::*
}

#----------------------------------
# Convenience routines
#----------------------------------

# Convert a (decimal) number into a bit vector string.  If dir = 0 (default),
# MSB is on the left.  if dir = 1, MSB is on the right.

proc irsim::bconvert {value bits {dir 0}} {
   set str ""
   for {set i 0} {$i < $bits} {incr i} {
      if {$dir == 0} {
	 set str "[expr {$value & 1}]$str"
      } else {
         append str [expr {$value & 1}]
      }
      set value [expr {$value >> 1}]
   }
   return $str
}

#----------------------------------
# Routines for the analyzer window
#----------------------------------

proc irsim::graphnode {nodename {row 0} {voffset 0}} {
   ::set glist [grid size .analyzer]
   ::set rtotal [lindex $glist 1]
   if {$rtotal >= $row} {

      irsim::analyzer -offset ${voffset} ${nodename}

      if {$rtotal == $row} {
         blt::stripchart .analyzer.row${row}
         grid rowconfigure .analyzer ${row} -weight 1
	 grid .analyzer.row${row} -row $row -column 0 -sticky news
         # Create the pen to use for unknown ('X', or 0.5 value) points
         .analyzer.row${row} pen create xPen -dashes {5 5} -symbol ""
      }

      ::set elist [.analyzer.row${row} element show]
      if {[lsearch $elist $nodename] < 0} {
         .analyzer.row${row} element create ${nodename} -xdata ${nodename}_time \
		-ydata ${nodename} -smooth step -symbol "" \
		-weights ${nodename} -styles {{xPen 0.25 0.75}}
      }

      if {[wm state .analyzer] == "withdrawn"} {wm deiconify .analyzer}
   }
}

#----------------------------------------------------------------------
# Cross-Application section
#----------------------------------------------------------------------

# Check namespaces for existence of other applications
set UsingMagic 0
set UsingXCircuit 0
set UsingNetgen 0
set nlist [namespace children]
foreach i $nlist {
   switch $i {
      ::magic { set UsingMagic 1 }
      ::xcircuit { set UsingXCircuit 1 }
      ::netgen { set UsingNetgen 1 }
   }
}

# Setup Magic assuming that the Tcl version is installed.

proc magic { args } {
   global CAD_HOME
   global argc
   global argv
   set magicscript [glob -nocomplain ${CAD_HOME}/lib/magic/tcl/magic.tcl]
   if { ${magicscript} == {} } {
      puts stderr "\"magic\" requires Tcl-based Magic version 7.2 or newer."
      puts stderr "Could not find script \"magic.tcl\".  If Magic is installed in a"
      puts stderr "place other than CAD_HOME (=${CAD_HOME}), use the command"
      puts stderr "\"source <path>/magic.tcl\"."
   } else {
      set argv $args
      set argc [llength $args]
      uplevel #0 source $magicscript
   }
}

# Setup Xcircuit assuming that the Tcl version is installed.

proc xcircuit { args } {
   global CAD_HOME
   global argc
   global argv
   set xcircscript [glob -nocomplain ${CAD_HOME}/lib/xcircuit*/xcircuit.tcl]
   if { ${xcircscript} == {} } {
      puts stderr "\"xcircuit\" requires Tcl-based XCircuit version 3.1 or newer."
      puts stderr "Could not find script \"xcircuit.tcl\".  If XCircuit is installed in a"
      puts stderr "place other than CAD_HOME (=${CAD_HOME}), use the command"
      puts stderr "\"source <path>/xcircuit.tcl\"."
   } else {
      # if there are multiple installed versions, choose the highest version.
      if {[llength $xcircscript] > 1} {
        set xcircscript [lindex [lsort -decreasing -dictionary $xcircscript] 0]
      }
      set argv $args
      set argc [llength $args]
      uplevel source $xcircscript
   }
}

# Setup Netgen assuming that the Tcl version is installed.

proc netgen { args } {
   global CAD_HOME
   global argc
   global argv
   set netgenscript [glob -nocomplain ${CAD_HOME}/lib/netgen/tcl/netgen.tcl]
   if { ${netgenscript} == {} } {
      puts stderr "\"netgen\" requires Tcl-based Netgen version 1.2 or newer."
      puts stderr "Could not find script \"netgen.tcl\".  If Netgen is installed in a"
      puts stderr "place other than CAD_HOME (=${CAD_HOME}), use the command"
      puts stderr "\"source <path>/netgen.tcl\"."
   } else {
      set argv $args
      set argc [llength $args]
      source $netgenscript
   }
}

#-------------------------------------------------------------------

set auto_noexec 1       ;# don't EVER call UNIX commands w/o "shell" in front

# Have we called irsim from tkcon or a clone thereof?  If so, set IrsimConsole

if {! $UsingMagic } {
   if {[lsearch [interp aliases] tkcon] != -1} {
      set IrsimConsole tkcon
      wm withdraw .
   }
}

pushnamespace irsim

#----------------------------------
# The analyzer window setup
#----------------------------------

toplevel .analyzer
wm geometry .analyzer 800x350+50+50
grid columnconfigure .analyzer 0 -weight 1
wm withdraw .analyzer

# Check for the presence of the "magic" namespace, and if it exists, set up
# handling between the two programs.

if { $UsingMagic } {

   array set magicnodes {}	;# initialize node dictionary
   set nodevalues {}

   # Callback routine which returns a node value to Tcl

   proc nodegetvalue { name value tval } {
      global nodevalues
      if {"${value}" != "t"} {lappend nodevalues ${value}}
   }

   # Callback which returns the time to Tcl

   proc timegetvalue { name value tval } {
      return $tval
   }

   # Magic-to-IRSIM node name conversion (to be done)

   proc magictoirsim { nodename } {
      return $nodename
   }

   # Watch the currently selected node in IRSIM

   proc watchnode { {wnode {} } {color white} } {
      global magicnodes
      global nodevalues

      if { $wnode == {} } {
         set wnode [magictoirsim [magic::getnode]]
	 if {$wnode == {} } { return }
      } else {
	 goto $wnode
         set wnode [magictoirsim $wnode]
      }
      set bvals [magic::box values]
      set fx [expr {([lindex $bvals 0] + [lindex $bvals 2]) / 2}]
      set fy [expr {([lindex $bvals 1] + [lindex $bvals 3]) / 2}]
      set nodevalues {}
      display tclproc nodegetvalue
      irsim::d ${wnode}
      set ival [lindex $nodevalues 0]
      display tclproc nodecallback
      magic::element add text irsim_$wnode $color $fx $fy "${wnode}=${ival}"
      irsim::w $wnode
      set magicnodes(${wnode}) irsim_${wnode}
   }

   # Stop watching a node

   proc unwatchnode { wnode } {
      global magicnodes
      magic::element delete irsim_$wnode
      unset magicnodes($wnode)
   }

   # Move a node to be centered in the current box

   proc movenode { wnode } {
      global magicnodes
      set bvals [magic::box values]
      set fx [expr {([lindex $bvals 0] + [lindex $bvals 2]) / 2}]
      set fy [expr {([lindex $bvals 1] + [lindex $bvals 3]) / 2}]
      magic::element configure $magicnodes(${wnode}) position $fx $fy
   }

   # Display the time in the magic window, positioned where the box is.

   proc watchtime {} {
      global magicnodes
      set bvals [magic::box values]
      set fx [expr {([lindex $bvals 0] + [lindex $bvals 2]) / 2}]
      set fy [expr {([lindex $bvals 1] + [lindex $bvals 3]) / 2}]
      display tclproc timegetvalue
      set tval [irsim::d]
      display tclproc nodecallback
      magic::element add text irsim_curtime white $fx $fy "time=${tval}ns"
      set magicnodes(curtime) irsim_curtime
   }

   # Stop watching the time

   proc unwatchtime {} {
      unwatchnode curtime
   }

   proc movetime {} {
      movenode curtime
   }

   # Search for a file in either the current directory or magic's
   # cell path.

   proc pathfile { filename } {
      global CAD_HOME
      set spath [join [list . [subst [magic::path cell]]]]
      foreach i $spath {
	 set rfile [glob -nocomplain ${i}/${filename}]
	 if { $rfile != {} } { return $rfile }
      }
      return {}
   }

   # Callback function for printing nodes and vectors

   proc nodecallback { name value tval } {
      global magicnodes
      if {"${value}" == "t"} {
	 if {[info exists magicnodes(curtime)]} {
	    magic::element configure $magicnodes(curtime) text "time=${tval}ns"
	    puts stdout "\n"
	 } else {
	    puts stdout "time=${tval}ns\n"
	 }
      } else {
	 if {[info exists magicnodes($name)]} {
	    magic::element configure $magicnodes($name) text "${name}=${value}"
	 } else {
	    puts -nonewline stdout "${name}=${value} "
	 }
      }
   }
   # Start irsim intelligently; assume basic parameters; extract and
   # do ext2sim if necessary.
   # To-do: guess the parameter file name from the extract style?

   proc irsim { { prmfile scmos100 } { cellname {} } } {
      if { $cellname == {} } {
	 # Get the current window's cellname
	 set curcell [ magic::cellname list window ]
	 set simfile [ pathfile ${curcell}.sim ]
	 if { $simfile == {} } {
	    set extfile [ pathfile ${curcell}.ext ]
	    if { $extfile == {} } {
	       magic::extract all
	    }
	    magic::exttosim
	 }
      }
      irsim::start $prmfile $simfile

      # Setup for callback functions displaying in magic.
      display tclproc nodecallback
   }

# If we are using magic, we return to the interpreter here;  IRSIM is
# started using the "irsim::start" command and the procedures just defined.

} else {

   if {[string range [wm title .] 0 3] == "wish"} {
      wm withdraw .
   }

#----------------------------------
# Parse command-line argument list
#----------------------------------

   if {[string length $argv] == 0} {
      irsim::start
   } else {
      set argafter {irsim::start}
      for {set i 0} {$i < $argc} {incr i 1} {
         set x [lindex $argv $i]
#
# Command-line argument handling goes here
# We have to handle all of irsim's command line arguments so we can
# figure out if a cell has been named for preloading.
#
# "-w" is reserved:  expects a "wrapper.tcl" program to run.
#
         switch -regexp -- $x {
            ^-w(rap)?(per)?$ {       ;# This regexp accepts -w, -wrap, and -wrapper
               source TCL_DIR/wrapper.tcl
            }
            -s -
            -n* {
               lappend argafter $x
            }
            default {
               lappend argafter $x
            }
         }
      }
      eval $argafter       ;# irsim::start ${argv}
   }
}

#----------------------------------------------------------------------------
# Irsim start function drops back to interpreter after initialization & setup
